
# AdaBoost

import model
from ..ailib import sign
import math


# for classification. classifier.err should be binary {0 = correct classification, 1 = incorrect classification} => err := I{y != G(x)}
class AdaBoost(model.Model):
    """AdaBoost algorithm.
    
    Given M (weak) classifiers, AdaBoost iteratively trains each one of them
    with respect to so far misclassified samples. For a formal representation,
    consult the respective :ref:`documentation <boosting-ada>`.
    
    As can be seen from the schematics, the classifiers are required to
    support weighted samples. The classifiers have to be linked to the instance
    of this class before training. This can be achieved through calls to the
    :func:`AdaBoost.addModel` method.
    
    """
    def __init__(self):
        self.models = []
        self.cWeight = []
    
    def addModel(self, m):
        """Add a trainable model to the set of boosted classifiers.
        
        The models are required to support weighted samples.
        
        :param m:       Model
        :type m:        :class:`Model`
        
        """
        self.models.append(m)
        self.cWeight.append( 0.0 ) # actually not required
    
    def fit(self, data):
        """Apply the AdaBoost algorithm on the presented data set.
        
        :param data:        Training data.
        :type data:         
        :returns:           self
        
        """
        N = len(data)
        
        # Initialize weights: wi = 1/n
        weights = [1.0/N for i in range(N)]
        
        self.cWeight = []
        for m in self.models:
            # a) Train a classifier c on the weighted training data
            m.fit(data, weights)
            
            # b) Compute error eb := sum( wi * I{yi != cb(x(i))}, i=1, n) / sum( wi, i=1, n)
            err = sum( [weights[i] * m.err(data[i]) for i in range(N) ] ) / sum(weights)
            
            # c) Compute voting weights: ab = log((1-eb) / eb)
            if err == 0.0:
                a = float('inf')
            else:
                a = math.log((1.0 - err) / err)
            self.cWeight.append( a )
            
            # d) Recompute weights: wi = wi * exp (ab * I{yi != cb(x(i))} )
            for i in range(N):
                weights[i] = weights[i] * math.exp(a * m.err(data[i]))
        return self

    def eval(self, x):
        """Evaluate the model at data point *x*.
        
        The outcome is the majority vote over the outcome of all classifiers
        w.r.t. their weight (note the :ref:`formal representation <boosting-ada-eval>`).
        """
        return sign(sum( [self.cWeight[m] * self.models[m].eval(x) for m in range(len(self.models))] ))


    def err(self, (x,y)):
        """Return the distance between the target *y* and the model
        prediction at the point *x*."""
        p = self.eval(x)
        if p == y: return 0.0
        return 1.0
    
    def __str__(self):
        m = []
        for i in range(len(self.models)):
            m.append( str(self.cWeight[i]) + ' * ' + str(self.models[i]) )
        return "AdaBoost(" + ' + '.join(m) + ")"
