from ailib import argmin, head, mean, snd
from ailib.sampling import rouletteWheelN
import kNN
import math
from math import exp
import random
import scipy

# TODO: Kmeans-Committee with guarantee that the initials are selected differently for each committee member.
# TODO: K-Means for gaussian mixtures


class Kmeans(kNN.kNN):
    """K-Means algorithm.
    
    This is an implementation of the probabilistic K-Means algorithm, not
    the often used Lloyd algorithm. The default evaluation method is to
    assign the sample to the closest cluster. Instead, the class label can
    be sampled from the distances, i.e. the closer to a cluster, the more
    probable this cluster label will be returned. If this behaviour is desired,
    use the :class:`Kmeans.Soft` mixin.
    
    There are several convergence criterions:
    
    - Risk is below a threshold (*absoluteRisk*).
    - risk decrease is below a threshold (*relativeRisk*).
    - Stop after a number of iterations (*iterCriterion*).
    - Assignments don't change (*centerCriterion*).
    
    Select one of those by assignment to *obj._converged*.
    
    """
    def __init__(self, numClusters, tol=1, maxIter=100, temp=1.0, dist=lambda x,y: norm(matrix(x)-matrix(y))**2):
        """
        :param numClusters: Number of clusters.
        :type numClusters: int
        :param tol: Threshold for the convergence criterion.
        :type tol: float
        :param maxIter: Limit to number of iterations.
        :type maxIter: int
        :param temp: Temperature.
        :type temp: float
        :param dist: Distance measurement
        :type dist: a -> a -> float
        """
        kNN.kNN.__init__(self, 1, dist)
        self.numClusters = numClusters
        self.tol = tol
        self.maxIter = maxIter
        self.temp = temp
        
        # No center's position changes more than threshold
        self.centerCriterion = lambda x,(Pp, yp),(P,y),w: not any(map(lambda i: w[i] * self.dist(yp[i],y[i]) > self.tol, range(self.numClusters)))	# center cirterion
        
        # Risk is below threshold
        self.absoluteRisk = lambda x,(Pp,yp),(P,y),w: self._risk(x,P,y,w) < self.tol
        
        # Risk decrease is below threshold
        self.relativeRisk = lambda x,(Pp,yp),(P,y),w: self._risk(x,Pp,yp,w) - self._risk(x,P,y,w) < self.tol

        # Wait for iterations to finish
        self.iterCriterion = lambda x,(Pp, yp),(P,y),w: False
        
        # set default
        self._converged = self.iterCriterion

    def _risk(self, x, P, y, w):
        """Risk of assignment: sum(w*dist(x,center(x)))/N."""
        N = head(P.shape)
        s = 0.0
        for i in range(N):
            clIdx = max(range(self.numClusters), key=lambda i: head(P[i].tolist()))
            s += w[i] * self.dist(x[i], y[clIdx])
        return s / N

    def fit(self, data, labels=None, weights=None, T=None):
        """
        
        :param data:        Training data.
        :type data:         [a]
        :param labels:      Unused.
        :type labels:       None
        :param weigths:     Unused.
        :type weights:      None
        :param T:           Temperature. If set, overwrites member.
        :type t:            float
        :returns:           self
        
        """
        if weights is None: weights = [1.0] * len(data)
        if T is None: T = self.temp
        # number of input points N
        N = len(data)
        # cluster centers; number of clusters K
        y = random.sample(data, self.numClusters)
        # assignment probabilities: P[NxK]
        P = scipy.matrix([[1.0/self.numClusters] * self.numClusters for j in range(N)])
        
        converged = False
        numIter = self.maxIter
        while not (numIter <= 0 or converged):
            prev = (P.copy(),y[:])
            
            # Update assignments (E-Step): P[i,a] = exp( -(x[i] - y[a])**2 / T) / sum( exp( -(x[i]-y[v])**2/T ), for all v )
            for i in range(N):
                for a in range(self.numClusters):
                    P[i,a] = math.exp( -weights[i] * self.dist(data[i], y[a]) / T) / sum( [math.exp( -weights[i] * self.dist(data[i],y[v]) / T ) for v in range(self.numClusters)] )
            
            # Update centers (M-Step): sum( P[i,a] * x[i], for all i) / sum( P[i,a], for all i )
            for a in range(self.numClusters):
                y[a] = sum( [P[i,a] * data[i] for i in range(N)]) / sum( [P[i,a] for i in range(N)] )

            converged = self._converged(data, prev, (P,y), weights)
            numIter -= 1
        
        self.data = [(y[i], i, 1.0) for i in range(self.numClusters)]
        return self

    class Soft:
        """Sample from clusters according to their distances."""
        def eval(self, x):
            """Return the class of *x*, sampled from the available clusters."""
            dist = [(exp(-self.dist(x, c) / self.temp),l) for (c,l,w) in self.data]
            weights = map(lambda (d,l): d, dist)
            return snd(rouletteWheelN(dist, weights, 1)[0])
