
from ailib import majorityVote, Format, snd
from ailib.sampling import rouletteWheelN
from model import Model
import random
from scipy import matrix
from scipy.linalg import norm

class kNN(Model):
    """k-Nearest Neighbour matching.
    
    A data point *x* is assigned to the same class as the most
    of its k nearest neighbours.
    
    if weights are given, the distance will be computed as
    
    >>> 1.0/(N*weight[i]) * dist(data[i],x)
    
    There are three evaluation types:
    
    - The most often appeared label (majority vote, the default).
    - Sampled from all neighbours with uniform probability (mixin Sampling).
    - Sampled from all neighbours with respect to their 
      distances (mixin WeightedSampling).
    
    The distance measurements
    
    - *distAbs*: absolute distance :math:`| x-y |`.
    - *distNorm*: norm :math:`\| x-y \|`.
    - *distSqNorm*: squared norm :math:`\| x-y \|^2`.
    
    are made available. Change the distance function by assignment to *obj.dist*
    For the lower ones, the data should be passed as list or a scipy.matrix.
    The default is distSqNorm.
    
    """
    def __init__(self, k, dist=lambda x,y: norm(matrix(x) - matrix(y))**2):
        """
        :param k:       Parameter which specifies the size of the neighbourhood.
        :type k:        int
        :param dist:    Distance measurement on data points.
        :type dist:     (Num b) => a -> a -> b
        """
        self.k = k
        self.dist = dist
        
        self.distAbs = lambda x,y: abs(x-y)
        self.distNorm = lambda x,y: norm(matrix(x) - matrix(y))
        self.distSqNorm = lambda x,y: norm(matrix(x) - matrix(y))**2
    
    def fit(self, data, labels, weights=None):
        """Set up the data points for later matching."""
        if weights is None: weights = [1.0] * len(data)
        self.data = Format.seperated.packed((data, labels, weights))
        return self
    
    def err(self, (x,y)):
        """:term:`0-1 loss function`."""
        e = self.eval(x)
        if e == y: return 0.0
        return 1.0
    
    def eval(self, x):
        """Return the label of the most frequent class of the
        k nearest neighbours of *x*.
        
        If there are two classes with the same number of occurrences
        in the k-neighbourhood of *x*, one of them is chosen at random
        (i.e. ties are broken randomly).
        """
        # distance of x to all data points
        dists = [(1.0 / w * self.dist(x, v), l) for (v,l,w) in self.data]
        dists.sort(key=lambda (d,l): d)
        best = dists[:self.k]
        
        # majority vote on labels
        return majorityVote(map(snd, best))

    class Sampling:
        """Mixin for sampling with uniform probabilities."""
        def eval(self, x):
            """Samples the label from all k nearest neighbours.
            
            All neighbours are equally likely to be drawn.
            """
            # distance of x to all data points
            dists = [(1.0 / w * self.dist(x, v), l) for (v,l,w) in self.data]
            dists.sort(key=lambda (d,l): d)
            best = dists[:self.k]
            return snd(random.choice(best))

    class WeightedSampling:
        """Mixin for sampling with respect to distances."""
        def eval(self, x):
            """Samples the label from all k nearest neighbours with respect to
            their distance.
            
            The sampling weights are inverse proportional to the distance:
            
            >>> w[i] = 1.0/(dist(data[i],x))
            
            with weights enabled
            
            >>> w[i] = N * weights[i] / dist(data[i], x)
            
            """
            # distance of x to all data points
            dists = [(1.0 / w * self.dist(x, v), l) for (v,l,w) in self.data]
            dists.sort(key=lambda (d,l): d)
            best = dists[:self.k]
            
            weights = map(lambda (d,l): 1.0/(d + 0.00001), best)
            return snd(rouletteWheelN(best, weights, 1)[0])
